import json
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator

from eunomia_core.enums.policy import ConditionOperator, PolicyEffect


class Condition(BaseModel):
    path: str = Field(
        ...,
        description="Path to the attribute in dot notation (e.g., 'attributes.role')",
    )
    operator: ConditionOperator = Field(..., description="Comparison operator")
    value: Any = Field(..., description="Value to compare against")

    model_config = ConfigDict(from_attributes=True)

    @field_validator("value", mode="before")
    @classmethod
    def parse_json(cls, v: Any) -> Any:
        if isinstance(v, str):
            try:
                return json.loads(v)
            except json.JSONDecodeError:
                return v
        return v


class Rule(BaseModel):
    name: str = Field(..., description="Name of the rule")
    effect: PolicyEffect = Field(..., description="Effect when the rule matches")
    principal_conditions: list[Condition] = Field(
        default_factory=list, description="Conditions applied to principal"
    )
    resource_conditions: list[Condition] = Field(
        default_factory=list, description="Conditions applied to resource"
    )
    actions: list[str] = Field(..., description="All actions evaluated by the rule")

    model_config = ConfigDict(from_attributes=True)

    @field_validator("actions", mode="before")
    @classmethod
    def parse_json(cls, v: list[str] | str) -> list[str]:
        if isinstance(v, str):
            return json.loads(v)
        return v


class Policy(BaseModel):
    version: str = Field("1.0", description="Version of the policy")
    name: str = Field(..., description="Name of the policy")
    description: Optional[str] = Field(
        None, description="Human-readable description of the policy"
    )
    rules: list[Rule] = Field(..., description="list of rules to evaluate")
    default_effect: PolicyEffect = Field(
        PolicyEffect.DENY, description="Default effect if no rules match"
    )

    model_config = ConfigDict(from_attributes=True, extra="forbid")


class PolicyEvaluationResult(BaseModel):
    effect: PolicyEffect = Field(..., description="The resulting effect")
    matched_rule: Optional[Rule] = Field(
        None, description="The rule that determined the effect, if any"
    )
    policy_name: str = Field(
        ..., description="The name of the policy that was evaluated"
    )



from .check import (
    CheckRequest,
    CheckResponse,
    EntityCheck,
    PrincipalCheck,
    ResourceCheck,
)
from .entity import (
    Attribute,
    AttributeInDb,
    EntityCreate,
    EntityInDb,
    EntityUpdate,
)
from .policy import Condition, Policy, PolicyEvaluationResult, Rule

__all__ = [
    "CheckRequest",
    "CheckResponse",
    "EntityCheck",
    "PrincipalCheck",
    "ResourceCheck",
    "Attribute",
    "AttributeInDb",
    "EntityCreate",
    "EntityInDb",
    "EntityUpdate",
    "Condition",
    "Policy",
    "PolicyEvaluationResult",
    "Rule",
]



from enum import Enum


class PolicyEffect(str, Enum):
    ALLOW = "allow"
    DENY = "deny"


class ConditionOperator(str, Enum):
    EQUALS = "equals"
    NOT_EQUALS = "not_equals"
    # String operators
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    STARTS_WITH = "startswith"
    ENDS_WITH = "endswith"
    # Number operators
    GREATER = "gt"
    GREATER_OR_EQUAL = "gte"
    LESS = "lt"
    LESS_OR_EQUAL = "lte"
    # List operators
    IN = "in"
    NOT_IN = "not_in"



import json
from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator

from eunomia_core.enums.entity import EntityType
from eunomia_core.utils import generate_uri


class Attribute(BaseModel):
    key: str = Field(..., description="Attribute key")
    value: Any = Field(..., description="Attribute value")

    @field_validator("value", mode="before")
    @classmethod
    def parse_json(cls, v: Any) -> Any:
        if isinstance(v, str):
            try:
                return json.loads(v)
            except json.JSONDecodeError:
                return v
        return v


class AttributeInDb(Attribute):
    updated_at: datetime = Field(
        description="Time when this attribute was last updated"
    )
    registered_at: datetime = Field(
        description="Time when this attribute was first registered"
    )

    model_config = ConfigDict(from_attributes=True)


class EntityBase(BaseModel):
    uri: str = Field(..., description="Unique identifier for the entity")
    attributes: list[Attribute] = Field(..., description="Entity attributes")
    type: EntityType = Field(..., description="Type of entity")

    @field_validator("attributes", mode="before")
    @classmethod
    def from_dict(cls, v: list[Attribute] | dict) -> list[Attribute]:
        if isinstance(v, dict):
            return [Attribute(key=key, value=value) for key, value in v.items()]
        return v

    @field_validator("attributes", mode="after")
    @classmethod
    def unique_attribute_keys(cls, v: list[Attribute]) -> list[Attribute]:
        keys = {}
        for i, attr in enumerate(v):
            if attr.key in keys:
                raise ValueError(
                    f"Duplicate attribute key: '{attr.key}' at positions {keys[attr.key]} and {i}"
                )
            keys[attr.key] = i
        return v


class EntityCreate(EntityBase):
    uri: Optional[str] = Field(
        default_factory=lambda: generate_uri(),
        description="Unique identifier for the entity, generated if not provided",
    )

    @field_validator("attributes", mode="before")
    @classmethod
    def at_least_one_attribute(cls, v: list[Attribute]) -> list[Attribute]:
        if not v:
            raise ValueError("At least one attribute must be provided")
        return v

    @field_validator("uri", mode="after")
    @classmethod
    def enforce_uri(cls, v: str | None) -> str:
        if v is None:
            return generate_uri()
        return v


class EntityUpdate(EntityBase):
    type: Optional[EntityType] = None  # type is not required for the attributes update

    @field_validator("attributes", mode="before")
    @classmethod
    def at_least_one_attribute(cls, v: list[Attribute]) -> list[Attribute]:
        if not v:
            raise ValueError("At least one attribute must be provided")
        return v


class EntityInDb(EntityBase):
    attributes: list[AttributeInDb] = Field(..., description="Entity attributes")
    registered_at: datetime = Field(description="Time when this entity was registered")

    model_config = ConfigDict(from_attributes=True)

    @property
    def attributes_dict(self) -> dict[str, Any]:
        return {attr.key: attr.value for attr in self.attributes}
